/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2021, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef headerfile_morm_modelwrapper
#define headerfile_morm_modelwrapper

#include <map>
#include <vector>
#include <list>
#include <string>


namespace morm
{
class Model;



class ModelWrapper
{
public:

	ModelWrapper()
	{
		auto_remove = true;
	}

	virtual ~ModelWrapper()
	{
		clear_childs();
	}

	virtual bool should_be_auto_removed()
	{
		return auto_remove;
	}

	virtual Model * get_model()
	{
		return nullptr;
	}

	virtual bool is_container()
	{
		return false;
	}

	virtual bool is_container_empty()
	{
		return true;
	}

	virtual void increment_iterator()
	{
	}

	virtual bool is_iterator_correct()
	{
		return false;
	}

	virtual void clear_childs()
	{
		for(auto & map_item : childs_map)
		{
			if( map_item.second->should_be_auto_removed() )
			{
				delete map_item.second;
				map_item.second = nullptr;
			}

			childs_map.clear();
		}
	}

	virtual void add_child(const std::wstring & child_name, ModelWrapper * models_base)
	{
		childs_map[child_name] = models_base;
	}

	virtual ModelWrapper * find_child(const std::wstring & child_name)
	{
		auto i = childs_map.find(child_name);

		if( i != childs_map.end() )
		{
			return i->second;
		}

		return nullptr;
	}

protected:

	bool auto_remove;
	std::map<std::wstring, ModelWrapper*> childs_map;


private:

	ModelWrapper(const ModelWrapper &)
	{
		auto_remove = false;
	}

	ModelWrapper & operator=(const ModelWrapper &)
	{
		return *this;
	}


};





class ModelWrapperModel : public ModelWrapper
{
public:

	ModelWrapperModel(Model * model)
	{
		this->model = model;
	}

	Model * get_model()
	{
		return model;
	}


protected:

	Model * model;

};






template<typename VectorType>
class ModelWrapperVector : public ModelWrapper
{
public:

	ModelWrapperVector(std::vector<VectorType> * container)
	{
		this->container = container;
		index = container->size();
	}

	bool is_container()
	{
		return true;
	}

	bool is_container_empty()
	{
		return container->empty();
	}

	void increment_iterator()
	{
		if( index >= container->size() )
		{
			index = 0;
		}
		else
		{
			index += 1;
		}
	}

	bool is_iterator_correct()
	{
		return index < container->size();
	}

	Model * get_model()
	{
		if( index < container->size() )
		{
			return &(*container)[index];
		}

		return nullptr;
	}


protected:

	std::vector<VectorType> * container;
	size_t index;


};






template<typename ListType>
class ModelWrapperList : public ModelWrapper
{
public:

	ModelWrapperList(std::list<ListType> * container)
	{
		this->container = container;
		iterator = container->end();
	}

	bool is_container()
	{
		return true;
	}

	bool is_container_empty()
	{
		return container->empty();
	}

	void increment_iterator()
	{
		if( iterator == container->end() )
		{
			iterator = container->begin();
		}
		else
		{
			++iterator;
		}
	}

	bool is_iterator_correct()
	{
		return iterator != container->end();
	}

	Model * get_model()
	{
		if( iterator != container->end() )
		{
			return &(*iterator);
		}

		return nullptr;
	}


protected:

	std::list<ListType> * container;
	typename std::list<ListType>::iterator iterator;

};



}

#endif
